#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>

/**
  * This is an implementation, in C, of the Longest Common Subsequence algorithm.
  * That is, given two strings A and B, this program will find the longest sequence
  * of letters that are common and ordered in A and B.
  *
  * There are only two reasons you are reading this:
  *   - you don't care what the algorithm is but you need a piece of code to do it
  *   - you're trying to understand the algorithm, and a piece of code might help
  * In either case, you should either read an entire chapter of an algorithms textbook
  * on the subject of dynamic programming, or you should consult a webpage that describes
  * this particular algorithm.   It is important, for example, that we use arrays of size
  * |A|+1 x |B|+1.
  *
  * This code is provided AS-IS.
  * You may use this code in any way you see fit, EXCEPT as the answer to a homework
  * problem or as part of a term project in which you were expected to arrive at this
  * code yourself.
  *
  * Copyright (C) 2005 Neil Jones.
  *
  */

/**
  * Takes two NULL-terminated strings.  Returns the longest common subsequence to both.
  */
char* LCS(char* X, char* Y);

#define NEITHER       0
#define UP            1
#define LEFT          2
#define UP_AND_LEFT   3

int main(int argc, char* argv[])
{
    if (argc != 3)
    {
        char *name = basename(argv[0]);
        printf("Usage: %s string1 string2\nE.g. %s ALTRUISTIC ALGORITHMIC\n", name, name);
        return 1;
    }
    printf("%s\n", LCS(argv[1],argv[2]));
    return 0;
}

/*
 * Note that this will allocate space for the result, and you will need
 * to free it yourself to avoid a memory leak.
 */
char* LCS(char* X, char* Y)
{
    int n = strlen(X);
    int m = strlen(Y);

    int** S;
    int** R;

    int i;
    int j;

    int pos;
    char* lcs;

    /*
     * Memory allocation.  Sigh.  What a quaint little idea. Nobody does this any more...
     * A much better way to do this is to allocate one big array for S (and one for R)
     * and then do your own pointer indexing.  However, that makes the card just slightly
     * harder to read, so I avoid it.
     */
    S = (int **)malloc( (n+1) * sizeof(int) );
    R = (int **)malloc( (n+1) * sizeof(int) );
    for(i = 0; i <= n; ++i)
    {
        S[i] = (int*) malloc( (m+1) * sizeof(int) );
        R[i] = (int*) malloc( (m+1) * sizeof(int) );
    }

    /* It is important to use <=, not <.  The next two for-loops are initialization */
    for(i = 0; i <= n; ++i)
    {
        S[i][0] = 0;
        R[i][0] = UP;
    }
    for(j = 0; j <= m; ++j)
    {
        S[0][j] = 0;
        R[0][j] = LEFT;
    }

    /* This is the main dynamic programming loop that computes the score and */
    /* backtracking arrays. */
    for(i = 1; i <= n; ++i)
    {
        for(j = 1; j <= m; ++j)
        {
            if( X[i-1] == Y[j-1] )
            {
                S[i][j] = S[i-1][j-1] + 1;
                R[i][j] = UP_AND_LEFT;
            }
            else
            {
                S[i][j] = S[i-1][j-1] + 0;
                R[i][j] = NEITHER;
            }
            if( S[i-1][j] >= S[i][j] )
            {
                S[i][j] = S[i-1][j];
                R[i][j] = UP;
            }
            if( S[i][j-1] >= S[i][j] )
            {
                S[i][j] = S[i][j-1];
                R[i][j] = LEFT;
            }
        }
    }
    // print S matrix
    printf("     ");
    for (j = 1; j <= m; j++)
        printf("%c ", Y[j-1]);
    printf("\n");
	for (i = 0; i <= n; i++)
	{
	    if (i) printf("%c ", X[i-1]);
	    else printf("  ");
		for (j = 0; j <= m; j++)
		{
			printf(" %d", S[i][j]);
		}
		printf("\n");
	}
	// print R matrix
    printf("     ");
    for (j = 1; j <= m; j++)
        printf("%c ", Y[j-1]);
    printf("\n");
	for (i = 0; i <= n; i++)
	{
	    if (i) printf("%c ", X[i-1]);
	    else printf("  ");
		for (j = 0; j <= m; j++)
		{
		    char dir;
			switch (R[i][j])
			{
			case UP_AND_LEFT:
			    dir = '\\';
			    break;
            case UP:
                dir = 'u';
                break;
            case LEFT:
                dir = 'l';
                break;
            default:
                dir = 'b';
			}
			printf(" %c", dir);
		}
		printf("\n");
	}
    /* The length of the longest substring is S[n][m] */
    i = n;
    j = m;
    pos = S[i][j];
    lcs = (char *) malloc( (pos+1) * sizeof(char) );

    lcs[pos--] = (char)0;

    /* Trace the backtracking matrix. */
    while( i > 0 || j > 0 )
    {
        if( R[i][j] == UP_AND_LEFT )
        {
            i--;
            j--;
            lcs[pos--] = X[i];
        }
        else if( R[i][j] == UP )
        {
            i--;
        }
        else if( R[i][j] == LEFT )
        {
            j--;
        }
    }
    for(i = 0; i <= n; ++i )
    {
        free(S[i]);
        free(R[i]);
    }
    free(S);
    free(R);

    return lcs;
}
